from .message import Message
from threading import Event


class MessageEvent:
    def __init__(self, call_back=None):
        self.event = Event()
        self.message = None
        self.call_back = call_back

    def trigger(self, message: Message):
        self.message = message
        self.event.set()
        if self.call_back:
            self.call_back(self.message)

    def wait(self, time_out: int = 0):
        time_out_sec = None
        if time_out:
            time_out_sec = float(time_out) / 1000
        if self.event.wait(timeout=time_out_sec):
            return self.message
        return None



import socket
from threading import Thread
from .connection import Connection
from .router import Router
from .message import Message
from .message_list import MessageList


class MessageServer:

    def __init__(self, ip: str = "0.0.0.0"):
        self.failed_messages = None
        self.messages = MessageList()
        self.router = Router()
        self.router.unrouted_message = self.__unrouted__
        self.connections = []
        self.thread = None
        self.server = socket.socket()
        self.ip = ip
        self.running = False
        self.thread = Thread(target=self.__proc__)
        self.thread.daemon = True
        self.allow_subscription = False
        self.subscriptions = []
        self.on_new_connection = None

    def broadcast(self, message: Message):
        to_remove = []
        for connection in self.connections:
            try:
                connection.send(message)
            except:
                to_remove.append(connection)
        for connection in to_remove:
            self.connections.remove(connection)

    def broadcast_subscribed(self, message: Message):
        to_remove = []
        for connection in self.subscriptions:
            try:
                connection.send(message)
            except:
                to_remove.append(connection)
        for connection in to_remove:
            self.subscriptions.remove(connection)

    def __unrouted__(self, message: Message):
        self.messages.append(message)

    def __subscribe_connection_fail__(self, message: Message):
        return False

    def __unsubscribe_connection_fail__(self, message: Message):
        return False

    def __subscribe_connection__(self, message: Message):
        if message._source in self.subscriptions:
            return False
        self.subscriptions.append(message._source)
        return True

    def __unsubscribe_connection__(self, message: Message):
        if message._source not in self.subscriptions:
            return False
        self.subscriptions.remove(message._source)
        return True

    def start(self, port: int):
        self.server.bind((self.ip, port))
        try:
            self.server.listen()
        except socket.error as err:
            return False
        self.server.settimeout(0.001)
        if self.allow_subscription:
            self.router.add_route("!subscribe", self.__subscribe_connection__)
            self.router.add_route("!unsubscribe", self.__unsubscribe_connection__)
        else:
            self.router.add_route("!subscribe", self.__subscribe_connection_fail__)
            self.router.add_route("!unsubscribe", self.__unsubscribe_connection_fail__)
        self.thread.start()
        while not self.running:
            pass
        return True

    def stop(self):
        if self.running:
            self.running = False
            for c in self.connections:
                c.close()
            self.thread.join()
            self.server.close()

    def __proc__(self):
        self.running = True
        while self.running:
            try:
                client, address = self.server.accept()
                if client:
                    client_connection = Connection(client, self.failed_messages)
                    self.connections.append(client_connection)
                    self.router.attend(client_connection)
                    if self.on_new_connection:
                        self.on_new_connection(client_connection)

            except socket.timeout:
                pass# no pending connecttions
            except Exception as e:
                print("Server: socked closed unexpectedly")
                self.running = False

    def join(self):
        if self.running:
            self.thread.join()

    def __del__(self):
        self.stop()

    def __bool__(self):
        if self.running:
            return True
        else:
            try:
                self.thread.join()
            finally:
                return False



from .message_list import MessageList
from .message import Message, MessageParts, MessagePart
import socket


class Connection:

    class State:
        Open = 1
        Close = 4

    def __init__(self, s: socket.socket, failed_message=None):
        self.socket = s
        self.socket.settimeout(0.001)
        self.failed_message = failed_message
        self.pending_messages = MessageList()
        self.state = None
        self.partials = dict()
        self.peek()

    def close(self):
        self.socket.close()
        self.state = Connection.State.Close

    def send(self, message: Message) -> bool:
        message_parts = MessageParts(message)
        for part in message_parts:
            message_str = str(part)
            message_bytes = message_str.encode()
            message_bytes += b'\x00'
            if self.socket.send(message_bytes) != len(message_bytes):
                return False
        return True

    def peek(self):
        try:
            data = self.socket.recv(1, socket.MSG_PEEK)
        except socket.timeout:
            self.state = Connection.State.Open
            return False
        except:
            self.state = Connection.State.Close
            return False
        if len(data) == 0:
            self.state = Connection.State.Close
            return False
        else:
            self.state = Connection.State.Open
            return True

    def receive(self):
        if not self.state == Connection.State.Open: #if the connection is not open
            return
        if self.pending_messages:
            return self.pending_messages.dequeue() #if there are pending messages retrun the oldest
        if not self.peek():
            return None #if there are no messages
        data = bytes()
        try:
            data = self.socket.recv(8192)
        except socket.timeout as e:
            pass
        except Exception as e:
            self.state = Connection.State.Close #if connection was closed from the other side
        else:
            if data:
                messages_str = data.decode().split('\x00')
                for message_str in messages_str:
                    if message_str:
                        try:
                            message_part = MessagePart.parse(message_str) # creates a new message instance
                            if message_part.parts <= 1:
                                self.pending_messages.queue(message_part.to_message())
                            else:
                                if message_part.id not in self.partials:
                                    self.partials[message_part.id] = MessageParts()
                                self.partials[message_part.id].append(message_part)
                                if self.partials[message_part.id].is_ready():
                                    self.pending_messages.queue(self.partials[message_part.id].join())
                                    del self.partials[message_part.id]
                        except:
                            if self.failed_message:
                                self.failed_message(message_str)
        return self.pending_messages.dequeue()

    def __bool__(self):
        self.peek()
        return self.state == Connection.State.Open



from json_cpp import JsonList
from .message import Message
from .message_event import MessageEvent


class MessageList(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Message)
        self.pending_responses = dict()

    def queue(self, message: Message):
        if message.id in self.pending_responses:
            self.pending_responses.pop(message.id).trigger(message)
        else:
            self.append(message)

    def add_message_event(self, request_id: str, event: MessageEvent):
        if request_id in self.pending_responses:
            return False
        self.pending_responses[request_id] = event
        return True

    def dequeue(self) -> Message:
        if len(self):
            message = self[0]
            del self[0]
            return message
        return None

    def contains(self, header: str) -> bool:
        for message in self:
            if message.header == header:
                return True
        return False

    def get_message(self, header: str) -> Message:
        for i in range(len(self)):
            if self[i].header == header:
                message = self[i]
                del self[i]
                return message
        return None

    def get_last_message(self, header: str) -> Message:
        message = None
        for i in range(len(self)):
            if self[i].header == header:
                message = self[i]
                del self[i]
        return message




from threading import Thread, Lock
import types
from .message import Message
from .message_event import MessageEvent
from .connection import Connection
from .util import check_type
import re
from json_cpp import JsonList, JsonObject


class Router:
    def __init__(self):
        self.routes = {}
        self.failed_message = None
        self.failed_route = None
        self.unrouted_message = None
        self.routing_count = 0
        self.pending_responses = dict()

    def add_message_event(self, request_id: str, event: MessageEvent):
        if request_id in self.pending_responses:
            return False
        self.pending_responses[request_id] = event
        return True

    def add_route(self, pattern: str, handler, body_type=None):
        check_type(handler, (types.FunctionType, types.MethodType), "incorrect type for handler")
        self.routes[pattern] = (handler, body_type)

    def get_manifest(self):
        manifest = JsonList()
        for pattern in self.routes.keys():
            (handler, body_type) = self.routes[pattern]
            if body_type:
                body_type_str = body_type.__name__
            else:
                body_type_str = ""
            manifest.append(JsonObject(route=pattern, input_type=body_type_str))
        return manifest

    def route(self, message: Message):
        responses = []
        check_type(message, Message, "incorrect type for message")
        if message.id in self.pending_responses:
            self.pending_responses[message.id].trigger(message)
            return responses
        if message.header == "!manifest":
            responses.append(self.get_manifest())
        for pattern in self.routes.keys():
            if re.search(pattern, message.header):
                (handler, body_type) = self.routes[pattern]
                try:
                    if body_type:
                        responses.append(handler(message.get_body(body_type)))
                    else:
                        responses.append(handler(message))
                except:
                    if self.failed_route:
                        self.failed_route(message)
        if not responses:
            if self.unrouted_message:
                self.unrouted_message(message)
        self.routing_count += 1
        return responses

    def attend(self, connection: Connection):
        RouterProcess.attend(connection, self)


class RouterProcess:
    __handler = None
    __mutex = Lock()

    @staticmethod
    def attend(connection: Connection, router: Router):
        RouterProcess.__mutex.acquire()
        if RouterProcess.__handler is None:
            RouterProcess.__handler = RouterProcess()
        RouterProcess.__handler.connections.append((connection, router))
        RouterProcess.__mutex.release()

    def __init__(self):
        if RouterProcess.__handler:
            raise Exception("ConnectionHandler is a singleton, use ConnectionHandler.handle")
        self.connections = list()
        self.incoming_messages_threads = []
        self.running = False
        self.thread = Thread(target=self.__process__)
        self.thread.daemon = True
        self.thread.start()
        while not self.running:
            pass
        RouterProcess.__handler = self

    @staticmethod
    def __incoming_message__(connection: Connection, router: Router, message: Message):
        message._source = connection
        responses = router.route(message)
        if responses:
            for response in responses:
                if isinstance(response, Message):
                    response.id = message.id
                    connection.send(response)
                elif isinstance(response, bool):
                    response_message = Message(message.header + "_response", "success" if response else "fail")
                    response_message.id = message.id
                    connection.send(response_message)
                else:
                    if response:
                        response_message = Message(message.header + "_response", str(response))
                        response_message.id = message.id
                        connection.send(response_message)

    def __process__(self):
        self.running = True
        while self.running:
            clean_up_required = []
            for index, (connection, router) in enumerate(self.connections):
                try:
                    if connection.state == Connection.State.Open:
                        message = connection.receive()
                        if message:
                            #RouterProcess.__incoming_message(connection, router, message)
                            Thread(target=RouterProcess.__incoming_message__, args=(connection, router, message)).start()
                    else:
                        clean_up_required.append(index)
                except:
                   clean_up_required.append(index)

            if clean_up_required:
                RouterProcess.__mutex.acquire()
                for failed_connection in clean_up_required:
                    try:
                        del self.connections[failed_connection]
                    except:
                        pass
                if len(self.connections) == 0:
                    RouterProcess.__handler = None
                    self.running = False
                RouterProcess.__mutex.release()

            to_remove = []
            while len(self.incoming_messages_threads) >= 10:
                for t in self.incoming_messages_threads:
                    if not t.is_alive():
                        to_remove.append(t)
                for t in to_remove:
                    self.incoming_messages_threads.remove(t)


